@@ -80,10 +80,10 @@ static void pthread_key_delete_helper(void *arg, pthread_descr th)
 
   if (!th->p_terminated) {
     /* pthread_exit() may try to free th->p_specific[idx1st] concurrently. */
-    __pthread_lock(THREAD_GETMEM(th, p_lock), self);
+    __pthread_lock(th->p_lock, self);
     if (th->p_specific[idx1st] != NULL)
       th->p_specific[idx1st][idx2nd] = NULL;
-    __pthread_unlock(THREAD_GETMEM(th, p_lock));
+    __pthread_unlock(th->p_lock);
   }
 }
 
@@ -208,40 +208,28 @@ void __pthread_destroy_specifics()
 
 /* Thread-specific data for libc. */
 
-static int
-libc_internal_tsd_set(enum __libc_tsd_key_t key, const void * pointer)
+int
+__pthread_internal_tsd_set (int key, const void * pointer)
 {
   pthread_descr self = thread_self();
 
   THREAD_SETMEM_NC(self, p_libc_specific[key], (void *) pointer);
   return 0;
 }
-int (*__libc_internal_tsd_set)(enum __libc_tsd_key_t key, const void * pointer)
-     = libc_internal_tsd_set;
 
-static void *
-libc_internal_tsd_get(enum __libc_tsd_key_t key)
+void *
+__pthread_internal_tsd_get (int key)
 {
   pthread_descr self = thread_self();
 
   return THREAD_GETMEM_NC(self, p_libc_specific[key]);
 }
-void * (*__libc_internal_tsd_get)(enum __libc_tsd_key_t key)
-     = libc_internal_tsd_get;
 
-static void ** __attribute__ ((__const__))
-libc_internal_tsd_address (enum __libc_tsd_key_t key)
+void ** __attribute__ ((__const__))
+__pthread_internal_tsd_address (int key)
 {
   pthread_descr self = thread_self();
   return &self->p_libc_specific[key];
 }
-void **(*const __libc_internal_tsd_address) (enum __libc_tsd_key_t key)
-     __THROW __attribute__ ((__const__)) = libc_internal_tsd_address;
 
 #endif
-
-int __libc_alloca_cutoff (size_t size)
-{
-  pthread_descr self = thread_self();
-  return size <= THREAD_GETMEM_NC(self, p_alloca_cutoff);
-}
